// splay.cpp
// implement module for splay tree classes
#include "splay.hpp"

//
// tree destroyer.  This is NOT the destructor for splay_tree_node --
// it is used by the splay_tree_root_class to delete the tree.
splay_tree_node_delete(splay_tree_node *current) {
//
//  A recursive version would kill the left subtree, then kill the right.
//
//  if(current == 0) return;
//  splay_tree_node_delete(current->left);
//  splay_tree_node_delete(current->right);
//  delete current;
//
// Since we've got pointers to track our tree traversal, we
// hand convert the recursive code to iterative code.
    splay_tree_node *tmp;
top:
    if(current->left != 0) { current = current->left; goto top; }
right:
    if(current->right != 0) { current = current->right; goto top; }
end:
    tmp = current->up;
    if(tmp != 0) {
        // if we're down a left subtree, we need to still do the right
        // subtree.
        if(tmp->left == current) { 
            delete current; current = tmp; goto right;
        // otherwise we've already been down the right subtree
        } else {
            delete current; current = tmp; goto end;
        }
    }
    delete current;
}

void
splay_tree_traverse(void (*func)(splay_tree_node *item),
                         splay_tree_node *current) {
    splay_tree_node *tmp;
top:
    if(current->left != 0) { current = current->left; goto top; }
dofunc:
    func(current);
    if(current->right != 0) { current = current->right; goto top; }
end:
    tmp = current->up;  // go up to the parent
    if(tmp != 0) {
        if(tmp->left == current) {
            current = tmp; goto dofunc; 
        } else {
            current = tmp; goto end;
        }
    }
}

inline splay_tree_node *
grandparent(splay_tree_node *item) { return item->up->up; }

inline splay_tree_node *
parent(splay_tree_node *item) { return item->up; }

//
// rotate the edge between a node and its right child.
inline void
rotate_left(splay_tree_node *item) {
    splay_tree_node *child = item->right, *p = parent(item);
    if(p != 0) {
        if(p->left == item) p->left = child; else p->right = child;
    }
    item->right = child->left; child->left = item;
    child->up = p; item->up = child;
    if(item->right != 0)
        item->right->up = item;
}

//
// rotate the edge between a node and its left child.
inline void
rotate_right(splay_tree_node *item) {
    splay_tree_node *child = item->left, *p = parent(item);
    if(p != 0) {
        if(p->right == item) p->right = child; else p->left = child;
    }
    item->left = child->right; child->right = item;
    child->up = p; item->up = child;
    if(item->left != 0)
            item->left->up = item;
}


//
// splay the tree around an item.  item should be in the current
// splay_tree_node, or the operation doesn't make much sense
void
splay_tree_root::splay(splay_tree_node *item) {
    while(parent(item) != 0) {
        if(item == parent(item)->left) {
            // zig case
            if(grandparent(item) == 0)
                rotate_right(parent(item));
            // zig-zig
            else if(parent(item) == grandparent(item)->left) {
                rotate_right(grandparent(item));
                rotate_right(parent(item));
            // zig-zag
            } else {
                rotate_right(parent(item));
                rotate_left(parent(item));
            }
        } else {
            // zig case
            if(grandparent(item) == 0)
                rotate_left(parent(item));
            // zig-zig
            else if(parent(item) == grandparent(item)->right) {
                rotate_left(grandparent(item));
                rotate_left(parent(item));
            // zig-zag
            } else {
                rotate_left(parent(item));
                rotate_right(parent(item));
            }
        }
    }
    //
    // item is now the new root of the tree, so we need to assign it
    // to root, because this is now out of date.
    root = item;
}

splay_tree_node *
splay_tree_root::access(splay_tree_node *item) {
    splay_tree_node *current = root, *last;
    int result;
    //
    // no sense searching an empty tree
    if(current == 0)
        return (splay_tree_node *)0;
    //
    // your basic binary search
    while((current != 0) && (result = compare(item,current)) != 0) {
        last = current;
        if(result < 0)
            current = current->left;
        else
            current = current->right;
    }
    //
    // if we find a node, splay there, else splay at last non-null node.
    splay(current == 0 ? last : current);
    return current;
}

void
splay_tree_root::insert(splay_tree_node *item) {
    splay_tree_node *current = root;
    int result;
    //
    // the inserted item has to start out with null children.
    item->left = item->right = 0;
    //
    // empty tree, just insert item at root.
    if(current == 0) {
        root = item; item->up = 0; return;
    }
    //
    // find location by binary search
    for(;;) {
        if((result = compare(item,current)) == 0) {
            // item already in tree -- replace it
            item->up = current->up;
            item->left = current->left;
            item->right = current->right;
            // tell item's children about their adoption
            current->left->up = item;
            current->right->up = item;
            //
            // patch parent pointers
            if(parent(current)->left == current)
                parent(current)->left = item;
            else
                parent(current)->right = item;
            item->up = parent(current);
            delete current;
            break;
        } else if(result < 0) {
            //
            // end of left subtree, put child here
            if(current->left == 0) {
                current->left = item; item->up = current; break;
            } else
                current = current->left;
        } else {
            //
            // end of right subtree, put child here
            if(current->right == 0) {
                current->right = item; item->up = current; break;
            } else
                current = current->right;
        }
    }
    //
    // always splay around item;
    splay(item);
}

splay_tree_node *
splay_tree_root::greatest(void) {
    splay_tree_node *current = root;

    if(current == 0)
        return 0;
    //
    // the greatest element is found by following the outermost right
    // edge of the tree.
    while(current->right != 0)
        current = current->right;
    return current;
}

splay_tree_node *
splay_tree_root::least(void) {
    splay_tree_node *current = root;

    if(current == 0)
        return 0;
    //
    // the least element is found by following the outermost left
    // edge of the tree.
    while(current->left != 0)
        current = current->left;
    return current;
}

void
splay_tree_root::remove(splay_tree_node *item) {
    splay_tree_node *current = access(item),*l,*r;
    //
    // if it isn't there, don't bother
    if(current == 0)
        return;
    //
    // if it is there, it's now the root. We want to join its
    // left and right subtrees to make the new tree.
    l = current->left; r = current->right;
    //
    // if we're deleting the least element, there isn't a left subtree
    if(l == 0) {
        root = r; r->up = 0;
    } else {
        //
        // the left subtree is the new root for the moment;
        root = l; l->up = 0;
        //
        // splay around the greatest element, this makes the right child
        // null, and we can then clip the right subtree of the deleted
        // item there;
        splay(greatest());
        root->right = r;
        if(r != 0)
            r->up = root;
    }
}
