//
// Class definitions for splay trees

class splay_tree_root;  // forward declaration

// splay_tree_node -- objects stored in the tree.
// This is subclassed to produce objects actually containing data.
class splay_tree_node {
    friend splay_tree_root; // root needs to know about structure,
                            // since it actually manipulate the tree
    friend splay_tree_node_delete(splay_tree_node *current);
    friend void splay_tree_traverse(void (*func)(splay_tree_node *item),
        splay_tree_node *current);
    friend splay_tree_node *grandparent(splay_tree_node *item);
    friend splay_tree_node *parent(splay_tree_node *item);
    friend void rotate_left(splay_tree_node *item);
    friend void rotate_right(splay_tree_node *item);
protected:
    splay_tree_node *left,*right,*up;
public:
    splay_tree_node(void) {}
};

// The comparison function for the particular subclass of a splay tree.
typedef int (*compare_function)(const splay_tree_node *item1, 
                   const splay_tree_node *item2);

// splay_tree_root -- class used to actually point at a splay tree.
// All of the access methods for splay trees are members of splay_tree_root.
class splay_tree_root {
protected:
    splay_tree_node *root;
private:
    compare_function compare;
    void splay(splay_tree_node *item);
public:
    splay_tree_root(compare_function cmp) { root = 0; compare = cmp; }
    ~splay_tree_root(void) { if(root != 0) splay_tree_node_delete(root); }
    void insert(splay_tree_node *item);
    void remove(splay_tree_node *item);
    splay_tree_node *access(splay_tree_node *item);
    splay_tree_node *greatest(void);
    splay_tree_node *least(void);
    void traverse(void (*func)(splay_tree_node *item)) { 
        splay_tree_traverse(func,root);
    }
};


//
// generic definitions of splay trees
#include <generic.hpp>

//
// This giant honker of a macro declares two new classes for a given
// existing data type -- a splay_tree_root class for that type, and
// a splay_tree_node type for that type.

#define make_splay_tree_subclass(type,comparison_function) \
/* forward definition of the root type */ \
class name2(type, _splay_tree_root) ; \
\
/* the node type.  It defines a copy constructor */ \
class name2(type, _splay_tree_node) : public splay_tree_node { \
public: \
    friend name2(type, _splay_tree_root); \
    type contained_data; \
    name2(type, _splay_tree_node) (type &data) { \
        contained_data = data; \
    } \
}; \
\
/* the root type. It defines basic operations in terms of the contained \
** data type \
*/ \
class name2(type, _splay_tree_root) : public splay_tree_root { \
public: \
    /* constructor passes pointer to comparison function to base type */ \
    name2(type, _splay_tree_root) (void) : (comparison_function) {}; \
    void insert(type &data) { \
        name2(type, _splay_tree_node) *tmp = \
            new name2(type, _splay_tree_node) (data); \
        splay_tree_root::insert(tmp); \
    } \
    void remove(type &data) { \
        name2(type, _splay_tree_node) tmp(data); \
        splay_tree_root::remove(&tmp); \
    } \
    type *access(type &data) { \
        splay_tree_node *rval; \
        name2(type, _splay_tree_node) tmp(data); \
        if((rval = splay_tree_root::access(&tmp)) == 0) \
            return 0; else return \
                &((name2(type, _splay_tree_node) *)rval)->contained_data; \
    } \
}
